@@ -19,12 +19,21 @@
 
 """The main module for the Orca screen reader."""
 
-__id__        = "$Id: orca.py,v 1.161 2006/10/05 16:31:03 wwalker Exp $"
-__version__   = "$Revision: 1.161 $"
-__date__      = "$Date: 2006/10/05 16:31:03 $"
+__id__        = "$Id: orca.py,v 1.176 2006/12/13 01:21:31 wwalker Exp $"
+__version__   = "$Revision: 1.176 $"
+__date__      = "$Date: 2006/12/13 01:21:31 $"
 __copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
 __license__   = "LGPL"
 
+# We're going to force the name of the app to "orca" so pygtk
+# will end up showing us as "orca" to the AT-SPI.  If we don't
+# do this, the name can end up being "-c".  See bug 364452 at
+# http://bugzilla.gnome.org/show_bug.cgi?id=364452 for more
+# information.
+#
+import sys
+sys.argv[0] = "orca"
+
 try:
     # This can fail due to gtk not being available.  We want to
     # be able to recover from that if possible.  The main driver
@@ -39,7 +48,6 @@ import getopt
 import os
 import signal
 import string
-import sys
 import time
 
 import atspi
@@ -69,6 +77,10 @@ from orca_i18n import _           # for gettext support
 #
 _userSettings = None
 
+# Command line options that override any other settings.
+#
+_commandLineSettings = {}
+
 ########################################################################
 #                                                                      #
 # METHODS FOR HANDLING PRESENTATION MANAGERS                           #
@@ -472,7 +484,8 @@ def _isModifierKey(event_string):
     #    event.]]]
 
     modifierKeys = [ 'Alt_L', 'Alt_R', 'Control_L', 'Control_R', \
-                     'Shift_L', 'Shift_R', 'Meta_L', 'Meta_R', "Insert" ]
+                     'Shift_L', 'Shift_R', 'Meta_L', 'Meta_R' ]
+    modifierKeys.extend(settings.orcaModifierKeys)
 
     reply = event_string in modifierKeys
     debug.println(debug.LEVEL_FINEST,
@@ -490,7 +503,8 @@ def _isLockingKey(event_string):
 
     lockingKeys = [ "Caps_Lock", "Num_Lock", "Scroll_Lock" ]
 
-    reply = event_string in lockingKeys
+    reply = event_string in lockingKeys \
+            and not event_string in settings.orcaModifierKeys
     debug.println(debug.LEVEL_FINEST,
                   "orca._echoLockingKey: returning: %s" % reply)
     return reply
@@ -632,6 +646,25 @@ def _keyEcho(event):
 
         speech.speak(event_string, voice)
 
+def _processKeyCaptured(event):
+    """Called when a new key event arrives and orca_state.capturingKeys=True.
+    (used for key bindings redefinition)
+    """
+
+    if event.type == 0:
+        if _isModifierKey(event.event_string) \
+               or event.event_string == "Return":
+            pass
+        elif event.event_string == "Escape":
+            orca_state.capturingKeys = False
+        else:
+            speech.speak(_("Key captured: %s. Press enter to confirm.") \
+                         % str(event.event_string))
+            orca_state.lastCapturedKey = event
+    else:
+        pass
+    return False
+
 def _processKeyboardEvent(event):
     """The primary key event handler for Orca.  Keeps track of various
     attributes, such as the lastInputEvent.  Also calls keyEcho as well
@@ -699,22 +732,29 @@ def _processKeyboardEvent(event):
     #
     consumed = False
     try:
-        consumed = _keyBindings.consumeKeyboardEvent(None, keyboardEvent)
-        if (not consumed) and (_currentPresentationManager >= 0):
-            consumed = _PRESENTATION_MANAGERS[_currentPresentationManager].\
-                       processKeyboardEvent(keyboardEvent)
-        if (not consumed) and settings.learnModeEnabled:
-            if keyboardEvent.type \
-                == atspi.Accessibility.KEY_PRESSED_EVENT:
-                # Check to see if there are localized words to be
-                # spoken for this key event.
-                #
-                braille.displayMessage(keyboardEvent.event_string)
-                event_string = keyboardEvent.event_string
-                if event_string in keynames.keynames:
-                    event_string = keynames.keynames[event_string]
-                speech.speak(event_string)
-            consumed = True
+        if orca_state.capturingKeys:
+            _processKeyCaptured(keyboardEvent)
+        else:
+            consumed = _keyBindings.consumeKeyboardEvent(None, keyboardEvent)
+            if (not consumed) and (_currentPresentationManager >= 0):
+                consumed = _PRESENTATION_MANAGERS[_currentPresentationManager].\
+                           processKeyboardEvent(keyboardEvent)
+            if (not consumed) and settings.learnModeEnabled:
+                if keyboardEvent.type \
+                    == atspi.Accessibility.KEY_PRESSED_EVENT:
+                    clickCount = util.getClickCount(orca_state.lastInputEvent,
+                                                keyboardEvent)
+                    if clickCount == 2:
+                        util.phoneticSpellCurrentItem(keyboardEvent.event_string)
+                    else:
+                        # Check to see if there are localized words to be
+                        # spoken for this key event.
+                        #
+                        braille.displayMessage(keyboardEvent.event_string)
+                        event_string = keyboardEvent.event_string
+                        event_string = keynames.getKeyName(event_string)
+                        speech.speak(event_string)
+                consumed = True
     except:
         debug.printException(debug.LEVEL_SEVERE)
 
@@ -794,6 +834,10 @@ def loadUserSettings(script=None, inputEvent=None):
     speech.shutdown()
     braille.shutdown()
     mag.shutdown()
+
+    if _currentPresentationManager >= 0:
+        _PRESENTATION_MANAGERS[_currentPresentationManager].deactivate()
+
     time.sleep(1)
 
     reloaded = False
@@ -809,6 +853,12 @@ def loadUserSettings(script=None, inputEvent=None):
         except ImportError:
             debug.printException(debug.LEVEL_FINEST)
 
+    # If any settings were added to the command line, they take
+    # precedence over everything else.
+    #
+    for key in _commandLineSettings:
+        settings.__dict__[key] = _commandLineSettings[key]
+
     if settings.enableSpeech:
         try:
             speech.init()
@@ -826,7 +876,7 @@ def loadUserSettings(script=None, inputEvent=None):
 
     if settings.enableBraille:
         try:
-            braille.init(_processBrailleEvent, 7)
+            braille.init(_processBrailleEvent, settings.tty)
         except:
             debug.printException(debug.LEVEL_WARNING)
             debug.println(debug.LEVEL_WARNING,
@@ -845,6 +895,13 @@ def loadUserSettings(script=None, inputEvent=None):
         debug.println(debug.LEVEL_CONFIGURATION,
                       "Magnification module has NOT been initialized.")
 
+    for keyName in settings.orcaModifierKeys:
+        if keyName == "Caps_Lock":
+            os.system('xmodmap -e "clear Lock"')
+
+    if _currentPresentationManager >= 0:
+        _PRESENTATION_MANAGERS[_currentPresentationManager].activate()
+
     _showMainWindowGUI()
 
     httpserver.init()
@@ -928,6 +985,22 @@ def _showQuitGUI(script=None, inputEvent=None):
 
     return True
 
+def _showFindGUI(script=None, inputEvent=None):
+    """Displays the user interace to perform an Orca Find.
+
+    Returns True to indicate the input event has been consumed.
+    """
+
+    try:
+        module = __import__(settings.findModule,
+                            globals(),
+                            locals(),
+                            [''])
+        module.showFindUI()
+    except:
+        debug.printException(debug.LEVEL_SEVERE)
+        pass
+
 # If True, this module has been initialized.
 #
 _initialized = False
@@ -1140,6 +1213,9 @@ def usage():
     print _("-s, --setup, --gui-setup     Set up user preferences")
     print _("-t, --text-setup             Set up user preferences (text version)")
     print _("-n, --no-setup               Skip set up of user preferences")
+    print _("-u, --user-prefs-dir=dirname Use alternate directory for user preferences")
+    print _("-e, --enable=[speech|braille|braille-monitor|magnifier|main-window] Force use of option")
+    print _("-d, --disable=[speech|braille|braille-monitor|magnifier|main-window] Prevent use of option")
     print _("-q, --quit                   Quits Orca (if shell script used)")
     print
     print _("If Orca has not been previously set up by the user, Orca\nwill automatically launch the preferences set up unless\nthe -n or --no-setup option is used.")
@@ -1154,6 +1230,8 @@ def main():
     an exit code of 0 means normal completion and an exit code of 50
     means Orca exited because of a hang."""
 
+    global _commandLineSettings
+
     # Method to call when we think something might be hung.
     #
     settings.timeoutCallback = timeout
@@ -1187,24 +1265,80 @@ def main():
     bypassSetup     = False
     setupRequested  = False
     showGUI         = False
+
+    # We hack a little here because the shell script to start orca can
+    # conflate all of command line arguments into one string, which is
+    # not what we want.  We detect this by seeing if the length of the
+    # argument list is 1.
+    #
+    arglist = sys.argv[1:]
+    if len(arglist) == 1:
+        arglist = arglist[0].split()
+
     try:
         # ? is for help
+        # e is for enabling a feature
+        # d is for disabling a feature
         # h is for help
+        # u is for alternate user preferences location
         # s is for setup
         # n is for no setup
         # t is for text setup
         # v is for version
         #
         opts, args = getopt.getopt(
-            sys.argv[1:],
-            "?stnv",
+            arglist,
+            "?stnvd:e:u:",
             ["help",
+             "user-prefs-dir=",
+             "enable=",
+             "disable=",
              "setup",
              "gui-setup",
              "text-setup",
              "no-setup",
              "version"])
         for opt, val in opts:
+            if opt in ("-u", "--user-prefs-dir"):
+                userPrefsDir = val.strip();
+                try:
+                    os.chdir(userPrefsDir)
+                    settings.userPrefsDir = userPrefsDir
+                except:
+                    debug.printException(debug.LEVEL_FINEST)
+
+            if opt in ("-e", "--enable"):
+                feature = val.strip()
+                if feature == "speech":
+                    _commandLineSettings["enableSpeech"] = True
+                elif feature == "braille":
+                    _commandLineSettings["enableBraille"] = True
+                elif feature == "braille-monitor":
+                    _commandLineSettings["enableBrailleMonitor"] = True
+                elif feature == "magnifier":
+                    _commandLineSettings["enableMagnifier"] = True
+                elif feature == "main-window":
+                    _commandLineSettings["showMainWindow"] = True
+                else:
+                    usage()
+                    os._exit(2)
+
+            if opt in ("-d", "--disable"):
+                feature = val.strip()
+                if feature == "speech":
+                    _commandLineSettings["enableSpeech"] = False
+                elif feature == "braille":
+                    _commandLineSettings["enableBraille"] = False
+                elif feature == "braille-monitor":
+                    _commandLineSettings["enableBrailleMonitor"] = False
+                elif feature == "magnifier":
+                    _commandLineSettings["enableMagnifier"] = False
+                elif feature == "main-window":
+                    _commandLineSettings["showMainWindow"] = False
+                else:
+                    usage()
+                    os._exit(2)
+
             if opt in ("-s", "--gui-setup", "--setup"):
                 setupRequested = True
                 showGUI = desktopRunning
@@ -1220,15 +1354,21 @@ def main():
                 print "Orca %s" % platform.version
                 os._exit(0)
     except:
+        debug.printException(debug.LEVEL_OFF)
         usage()
         os._exit(2)
 
     # Do not run Orca if accessibility has not been enabled.
+    # We do allow, however, one to force Orca to run via the
+    # "-n" switch.  The main reason is so that things such
+    # as accessible login can work -- in those cases, the gconf
+    # setting is typically not set since the gdm user does not
+    # have a home.
     #
     import commands
     a11yEnabled = commands.getoutput(\
         "gconftool-2 --get /desktop/gnome/interface/accessibility")
-    if a11yEnabled != "true":
+    if (not bypassSetup) and (a11yEnabled != "true"):
         _showPreferencesConsole()
         abort()
 
@@ -1241,7 +1381,7 @@ def main():
         print "environment variable has been set."
         return 1
 
-    userprefs = os.path.join(os.environ["HOME"], ".orca")
+    userprefs = settings.userPrefsDir
     sys.path.insert(0, userprefs)
     sys.path.insert(0, '') # current directory
 
