@@ -19,9 +19,9 @@
 
 """Custom script for Evolution."""
 
-__id__        = "$Id: Evolution.py,v 1.70 2006/11/01 17:13:37 richb Exp $"
-__version__   = "$Revision: 1.70 $"
-__date__      = "$Date: 2006/11/01 17:13:37 $"
+__id__        = "$Id: Evolution.py,v 1.75 2006/12/08 17:12:48 lmonsanto Exp $"
+__version__   = "$Revision: 1.75 $"
+__date__      = "$Date: 2006/12/08 17:12:48 $"
 __copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
 __license__   = "LGPL"
 
@@ -36,6 +36,7 @@ import orca.orca_state as orca_state
 import orca.speech as speech
 import orca.settings as settings
 import orca.util as util
+import orca.chnames as chnames
 
 from orca.orca_i18n import _ # for gettext support
 
@@ -185,7 +186,8 @@ class Script(default.Script):
             if label.text:
                 charAttributes = label.text.getAttributes(0)
                 if charAttributes[0]:
-                    charDict = self.textAttrsToDictionary(charAttributes[0])
+                    [charKeys, charDict] = \
+                        self.textAttrsToDictionary(charAttributes[0])
                     weight = charDict.get('weight')
                     if weight and weight == '800':
                         text = util.getDisplayedText(label)
@@ -383,13 +385,23 @@ class Script(default.Script):
                           + "current message pane: " \
                           + "individual lines of text.")
 
-            result = util.getTextLineAtCaret(event.source)
-            braille.displayMessage(result[0])
+            [string, caretOffset, startOffset] = \
+                util.getTextLineAtCaret(event.source)
+            braille.displayMessage(string)
             if settings.enableSpeechIndentation:
-                self.speakTextIndentation(event.source, result[0])
-            line = util.adjustForRepeats(result[0])
-            speech.speak(line)
+                self.speakTextIndentation(event.source, string)
+            line = util.adjustForRepeats(string)
+                
+            if self.speakNewLine(event.source):
+                speech.speak(chnames.getCharacterName("\n"), None, False)
+ 
+            if self.speakBlankLine(event.source):
+                speech.speak(_("blank"), None, False)
+            else:
+                speech.speak(line, None, False)
+
             return
+        
 
         # 2) Mail view: current message pane: "standard" mail header lines.
         #
@@ -436,8 +448,9 @@ class Script(default.Script):
 
                     if cell.role == rolenames.ROLE_TEXT:
                         regions.append(braille.Text(cell))
-                        result = util.getTextLineAtCaret(cell)
-                        utterances.append(result[0])
+                        [string, caretOffset, startOffset] = \
+                            util.getTextLineAtCaret(cell)
+                        utterances.append(string)
 
                 braille.displayRegions([regions, regions[0]])
                 speech.speakUtterances(utterances)
@@ -566,7 +579,7 @@ class Script(default.Script):
                             if speakAll or (column == i):
                                 speech.speakUtterances(utterances)
 
-                        # Speak/braille the table cell. 
+                        # Speak/braille the table cell.
                         #
                         # If this cell has a column header of "Status",
                         # then speak/braille "read".
@@ -853,7 +866,15 @@ class Script(default.Script):
                           + "compose window: message area.")
 
             self.message_panel = event.source.parent.parent
+            
+            if self.speakNewLine(event.source):
+                speech.speak(chnames.getCharacterName("\n"), None, False)
 
+
+            if self.speakBlankLine(event.source):
+                 speech.speak(_("blank"), None, False)
+
+                
         # 9) Spell Checking Dialog
         #
         # This works in conjunction with code in section 8). Check to see if
@@ -982,6 +1003,86 @@ class Script(default.Script):
         default.Script.locusOfFocusChanged(self, event,
                                            oldLocusOfFocus, newLocusOfFocus)
 
+    def speakNewLine(self, obj):
+        """Returns True if a newline should be spoken.
+           Otherwise, returns False.
+        """
+
+        # Get the the AccessibleText interrface.
+        text = obj.text
+        if not text:
+            return False
+
+        # Was a left or right-arrow key pressed?
+        if not (orca_state.lastInputEvent and \
+                orca_state.lastInputEvent.__dict__.has_key("event_string")):
+            return False
+        
+        lastKey = orca_state.lastInputEvent.event_string
+        if lastKey != "Left" and lastKey != "Right":
+            return False
+
+        # Was a control key pressed?
+        mods = orca_state.lastInputEvent.modifiers
+        isControlKey = mods & (1 << atspi.Accessibility.MODIFIER_CONTROL)
+        
+        # Get the line containing the caret
+        caretOffset = text.caretOffset
+        line = text.getTextAtOffset(caretOffset, \
+            atspi.Accessibility.TEXT_BOUNDARY_LINE_START)
+        lineStart = line[1]
+        lineEnd = line[2]
+        
+        if isControlKey:  # control-right-arrow or control-left-arrow
+
+            # Get the word containing the caret.
+            word = text.getTextAtOffset(caretOffset, \
+                atspi.Accessibility.TEXT_BOUNDARY_WORD_START)
+            wordStart = word[1]
+            wordEnd = word[2]
+            
+            if lastKey == "Right":
+                if wordStart == lineStart:
+                    return True
+            else: 
+                if wordEnd == lineEnd:
+                    return True
+                
+        else:  # right arrow or left arrow
+            
+            if lastKey == "Right":
+                if caretOffset == lineStart:
+                    return True
+            else: 
+                if caretOffset == lineEnd:
+                    return True
+
+        return False
+    def speakBlankLine(self, obj):
+        """Returns True if a blank line should be spoken.
+        Otherwise, returns False.
+        """
+        
+        # Get the the AccessibleText interrface. 
+        text = obj.text
+        if not text:
+            return False
+
+        # Get the line containing the caret
+        caretOffset = text.caretOffset
+        line = text.getTextAtOffset(caretOffset, \
+            atspi.Accessibility.TEXT_BOUNDARY_LINE_START)
+
+        debug.println(debug.LEVEL_FINEST, "speakBlankLine: start=%d, end=%d, line=<%s>" % (line[1], line[2], line[0]))
+
+        # If this is a blank line, announce it if the user requested
+        # that blank lines be spoken.
+        if line[1] == 0 and line[2] == 0:
+            return settings.speakBlankLines
+        else:
+            return False
+
+        
     def onStateChanged(self, event):
         """Called whenever an object's state changes.  We are only
         interested in "object:state-changed:showing" events for any
