@@ -19,19 +19,21 @@
 
 """Custom script for StarOffice and OpenOffice."""
 
-__id__        = "$Id: StarOffice.py,v 1.62 2006/10/29 22:50:12 mpeders Exp $"
-__version__   = "$Revision: 1.62 $"
-__date__      = "$Date: 2006/10/29 22:50:12 $"
+__id__        = "$Id: StarOffice.py,v 1.73 2006/12/14 16:48:08 richb Exp $"
+__version__   = "$Revision: 1.73 $"
+__date__      = "$Date: 2006/12/14 16:48:08 $"
 __copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
 __license__   = "LGPL"
 
 import orca.debug as debug
 import orca.atspi as atspi
+import orca.chnames as chnames
 import orca.default as default
 import orca.input_event as input_event
 import orca.rolenames as rolenames
 import orca.braille as braille
 import orca.braillegenerator as braillegenerator
+import orca.orca as orca
 import orca.orca_state as orca_state
 import orca.speech as speech
 import orca.speechgenerator as speechgenerator
@@ -143,6 +145,7 @@ def isSpreadSheetCell(obj):
     Returns True if this is a table cell, False otherwise.
     """
 
+    found = False
     rolesList = [rolenames.ROLE_TABLE_CELL, \
                  rolenames.ROLE_TABLE, \
                  rolenames.ROLE_UNKNOWN, \
@@ -151,7 +154,19 @@ def isSpreadSheetCell(obj):
                  rolenames.ROLE_ROOT_PANE, \
                  rolenames.ROLE_FRAME, \
                  rolenames.ROLE_APPLICATION]
-    return util.isDesiredFocusedItem(obj, rolesList)
+    if util.isDesiredFocusedItem(obj, rolesList):
+        # We've found a table cell with the correct hierarchy. Now check
+        # that we are in a spreadsheet as opposed to the writer application.
+        # See bug #382408.
+        #
+        current = obj.parent
+        while current.role != rolenames.ROLE_APPLICATION:
+            if current.role == rolenames.ROLE_FRAME and \
+               (current.name and current.name.endswith(_("Calc"))):
+                found = True
+            current = current.parent
+
+    return found
 
 class BrailleGenerator(braillegenerator.BrailleGenerator):
     """Overrides _getBrailleRegionsForTableCellRow so that , when we are
@@ -184,8 +199,8 @@ class BrailleGenerator(braillegenerator.BrailleGenerator):
         if isSpreadSheetCell(obj):
 
             # Check to see if this spread sheet cell has either a dynamic
-            # column heading or row heading (or both) associated with it. 
-            # If it does, then braille those first before brailling the 
+            # column heading or row heading (or both) associated with it.
+            # If it does, then braille those first before brailling the
             # cell contents.
             #
             table = getCalcTable(obj)
@@ -305,13 +320,31 @@ class BrailleGenerator(braillegenerator.BrailleGenerator):
             return [regions, componentRegion]
 
         else:
-            brailleGen = braillegenerator.BrailleGenerator
-            regions = brailleGen._getBrailleRegionsForTableCell(self, obj)
+            # Check to see how many children this table cell has. If it's
+            # just one (or none), then pass it on to the superclass to be
+            # processed.
+            #
+            # If it's more than one, then get the braille regions for each 
+            # child, and call this method again.
+            #
+            if obj.childCount <= 1:
+                brailleGen = braillegenerator.BrailleGenerator
+                regions = brailleGen._getBrailleRegionsForTableCell(self, obj)
+            else:
+                regions = []
+                for i in range(0, obj.childCount):
+                    child = obj.child(i)
+                    [cellRegions, focusRegion] = \
+                                self._getBrailleRegionsForTableCell(child)
+                    regions.extend(cellRegions)
+                return [regions, focusRegion]
 
         return regions
 
 class SpeechGenerator(speechgenerator.SpeechGenerator):
-    """Overrides _getSpeechForTableCellRow so that , when we are in a
+    """Overrides _getSpeechForComboBox so that we can provide a name for
+    the Calc Name combo box.
+    Overrides _getSpeechForTableCellRow so that , when we are in a
     spread sheet, we can speak the dynamic row and column headers
     (assuming they are set).
     Overrides _getSpeechForTableCell so that, when we are in a spread
@@ -321,6 +354,45 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
     def __init__(self, script):
         speechgenerator.SpeechGenerator.__init__(self, script)
 
+    def _getSpeechForComboBox(self, obj, already_focused):
+        """Get the speech for a combo box. If the combo box already has focus,
+        then only the selection is spoken.
+        Also provides a name for the OOo Calc Name combo box. This name is
+        provided in clause 5) of locusOfFocusChanged() below.
+
+        Arguments:
+        - obj: the combo box
+        - already_focused: False if object just received focus
+
+        Returns a list of utterances to be spoken for the object.
+        """
+
+        utterances = []
+
+        if not already_focused:
+            label = self._getSpeechForObjectLabel(obj)
+            if not label:
+                label = [ obj.name ]
+            utterances.extend(label)
+        else:
+            label = None
+
+        name = self._getSpeechForObjectName(obj)
+        if name != label:
+            utterances.extend(name)
+
+        if not already_focused:
+            utterances.extend(self._getSpeechForObjectRole(obj))
+
+        utterances.extend(self._getSpeechForObjectAvailability(obj))
+
+        self._debugGenerator("_getSpeechForComboBox",
+                             obj,
+                             already_focused,
+                             utterances)
+
+        return utterances
+
     def _getSpeechForTableCellRow(self, obj, already_focused):
         """Get the speech for a table cell row or a single table cell
         if settings.readTableCellRow is False. If this isn't inside a
@@ -437,9 +509,23 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
             nameList = obj.name.split()
             utterances.append(nameList[1])
         else:
-            speechGen = speechgenerator.SpeechGenerator
-            utterances = speechGen._getSpeechForTableCell(self, obj,
-                                                      already_focused)
+            # Check to see how many children this table cell has. If it's
+            # just one (or none), then pass it on to the superclass to be 
+            # processed.
+            #
+            # If it's more than one, then get the speech for each child,
+            # and call this method again.
+            #
+            if obj.childCount <= 1:
+                speechGen = speechgenerator.SpeechGenerator
+                utterances = speechGen._getSpeechForTableCell(self, obj,
+                                                        already_focused)
+            else:
+                utterances = []
+                for i in range(0, obj.childCount):
+                    child = obj.child(i)
+                    utterances.extend(self._getSpeechForTableCell(child,
+                                                        already_focused))
 
         return utterances
 
@@ -485,7 +571,7 @@ class Script(default.Script):
         # Used to determine whether the caret has moved to a new paragraph.
         #
         self.currentParagraph = None
-        
+
 
     def getBrailleGenerator(self):
         """Returns the braille generator for this script.
@@ -555,6 +641,41 @@ class Script(default.Script):
 
         return keyBindings
 
+    def checkForTableBoundry(self, oldFocus, newFocus):
+        """Check to see if we've entered or left a table.
+        When entering a table, announce the table dimensions.
+        When leaving a table, announce that the table has been exited.
+
+        Arguments:
+        - oldFocus: Accessible that is the old locus of focus
+        - newFocus: Accessible that is the new locus of focus
+        """
+
+        if oldFocus == None or newFocus == None:
+            return
+
+        oldFocusIsTable = None 
+        while oldFocus.role != rolenames.ROLE_APPLICATION:
+            if oldFocus.role == rolenames.ROLE_TABLE:
+                oldFocusIsTable = oldFocus
+                break
+            oldFocus = oldFocus.parent 
+
+        newFocusIsTable = None
+        while newFocus.role != rolenames.ROLE_APPLICATION:
+            if newFocus.role == rolenames.ROLE_TABLE:
+                newFocusIsTable = newFocus
+                break
+            newFocus = newFocus.parent
+
+        if oldFocusIsTable == None and newFocusIsTable != None:
+            rows = newFocusIsTable.table.nRows
+            columns = newFocusIsTable.table.nColumns
+            line = _("table with %d rows and %d columns.") % (rows, columns)
+            speech.speak(line)
+        elif oldFocusIsTable != None and newFocusIsTable == None:
+            speech.speak(_("leaving table."))
+
     def speakInputLine(self, inputEvent):
         """Speak the contents of the spread sheet input line (assuming we
         have a handle to it - generated when we first focus on a spread
@@ -617,7 +738,7 @@ class Script(default.Script):
         return column
 
     def setDynamicColumnHeaders(self, inputEvent):
-        """Set the row for the dynamic header columns to use when speaking 
+        """Set the row for the dynamic header columns to use when speaking
         calc cell entries. In order to set the row, the user should first set
         focus to the row that they wish to define and then press Insert-r.
 
@@ -679,9 +800,9 @@ class Script(default.Script):
         return res
 
     def setDynamicRowHeaders(self, inputEvent):
-        """Set the column for the dynamic header rows to use when speaking 
-        calc cell entries. In order to set the column, the user should first 
-        set focus to the column that they wish to define and then press 
+        """Set the column for the dynamic header rows to use when speaking
+        calc cell entries. In order to set the column, the user should first
+        set focus to the column that they wish to define and then press
         Insert-c.
 
         Once the user has defined the column, it will be used to first speak
@@ -889,10 +1010,12 @@ class Script(default.Script):
             self.speakSetupLabel(allLabels[i])
 
     # This method tries to detect and handle the following cases:
+    # 0) Writer: find command.
     # 1) Writer: text paragraph.
     # 2) Writer: spell checking dialog.
     # 3) Welcome to StarOffice dialog.
     # 4) Calc: cell editor.
+    # 5) Calc: name box.
 
     def locusOfFocusChanged(self, event, oldLocusOfFocus, newLocusOfFocus):
         """Called when the visual object with focus changes.
@@ -912,6 +1035,17 @@ class Script(default.Script):
 
         # util.printAncestry(event.source)
 
+        # 0) Writer: find command
+        #
+        # Check to see if this is this is for the find command. See 
+        # comment #18 of bug #354463.
+        #
+        if self.findCommandRun and \
+           event.type == "object:state-changed:focused":
+            self.findCommandRun = False
+            self.find()
+            return
+
         # 1) Writer: text paragraph.
         #
         # We need to handle two things here:
@@ -921,8 +1055,8 @@ class Script(default.Script):
         # paragraph, then we've just created the first new text document
         # in Writer. Announce it by doing a "where am I".
         #
-        # Also, when the focus is on a paragraph in the Document view of 
-        # the Writer, then just speak/braille the current line (rather than 
+        # Also, when the focus is on a paragraph in the Document view of
+        # the Writer, then just speak/braille the current line (rather than
         # speaking a bogus initial "paragraph" utterance as well).
 
         rolesList = [rolenames.ROLE_PARAGRAPH, \
@@ -1115,6 +1249,28 @@ class Script(default.Script):
                           + "Calc: cell editor.")
             return
 
+        # 5) Calc: name box
+        #
+        # Check to see if the focus has just moved to the Name Box combo
+        # box in Calc. If so, then replace the non-existent name with a
+        # simple one before falling through and calling the default 
+        # locusOfFocusChanged method, which in turn will result in our
+        # _getSpeechForComboBox() method being called.
+        #
+        rolesList = [rolenames.ROLE_COMBO_BOX, \
+                     rolenames.ROLE_TOOL_BAR, \
+                     rolenames.ROLE_PANEL, \
+                     rolenames.ROLE_ROOT_PANE, \
+                     rolenames.ROLE_FRAME, \
+                     rolenames.ROLE_APPLICATION]
+
+        if util.isDesiredFocusedItem(event.source, rolesList) \
+            and (not event.source.name or len(event.source.name) == 0):
+            debug.println(self.debugLevel, "StarOffice.locusOfFocusChanged - " \
+                          + "Calc: name box.")
+
+            event.source.name = _("Move to cell")
+
         # Pass the event onto the parent class to be handled in the default way.
 
         default.Script.locusOfFocusChanged(self, event,
@@ -1212,6 +1368,35 @@ class Script(default.Script):
         default.Script.onNameChanged(self, event)
 
 
+    def onFocus(self, event):
+        """Called whenever an object gets focus. Overridden in this script 
+        so that we can adjust "focus:" events for children of a combo-box 
+        to just set the focus to the combo box. This is needed to help 
+        reduce the verbosity of focusing on the Calc Name combo box (see 
+        bug #364407).
+
+        Arguments:
+        - event: the Event
+        """
+
+        if event.source.parent.role == rolenames.ROLE_COMBO_BOX:
+            orca.setLocusOfFocus(None, event.source.parent, False)
+            return
+
+        # If we are FOCUSED on a paragraph inside a table cell (in Writer),
+        # then speak/braille that parent table cell (see bug #382415).
+        #
+        if event.source.role == rolenames.ROLE_PARAGRAPH and \
+           event.source.parent.role == rolenames.ROLE_TABLE_CELL and \
+           event.source.state.count(atspi.Accessibility.STATE_FOCUSED):
+            if self.lastCell != event.source.parent:
+                default.Script.locusOfFocusChanged(self, event,
+                                                   None, event.source.parent)
+                self.lastCell = event.source.parent
+            return
+
+        default.Script.onFocus(self, event)
+
     def onStateChanged(self, event):
         """Called whenever an object's state changes.
 
@@ -1219,15 +1404,76 @@ class Script(default.Script):
         - event: the Event
         """
 
-        # Two focus changed events are received when the caret
-        # moves to a new paragraph. Ignore the second one.
+        # If this is a state change "focused" event that we care about, and 
+        # we are in Writer, check to see if we are entering or leaving a table.
         #
+        if event.type == "object:state-changed:focused" and event.detail1 == 1:
+            current = event.source.parent
+            while current.role != rolenames.ROLE_APPLICATION:
+                if current.role == rolenames.ROLE_FRAME and \
+                   (current.name and current.name.endswith(_("Writer"))):
+                    self.checkForTableBoundry(orca_state.locusOfFocus, 
+                                              event.source)
+                    break
+                current = current.parent
+
+        # Prevent  "object:state-changed:active" events from activating
+        # the find operation. See comment #18 of bug #354463.
+        #
+        if event.type == "object:state-changed:active":
+            if self.findCommandRun:
+                return
+
+        # If we are FOCUSED on a paragraph inside a table cell (in Writer),
+        # then speak/braille that parent table cell (see bug #382415).
+        #
+        if event.type == "object:state-changed:focused" and \
+           event.source.role == rolenames.ROLE_PARAGRAPH and \
+           event.source.parent.role == rolenames.ROLE_TABLE_CELL and \
+           event.detail1 == 1 and \
+           event.source.state.count(atspi.Accessibility.STATE_FOCUSED):
+
+            # Check to see if the last input event was "Up" or "Down".
+            # If it was, and we are in the same table cell as last time,
+            # and if that table cell has more than one child, then just
+            # get the speech for that single child, otherwise speak/braille
+            # the parent table cell.
+            #
+            event_string = orca_state.lastInputEvent.event_string
+            if (event_string == "Up" or event_string == "Down") and \
+               event.source.parent == self.lastCell and \
+               event.source.parent.childCount > 1:
+                default.Script.locusOfFocusChanged(self, event,
+                                                   None, event.source)
+            else:
+                default.Script.locusOfFocusChanged(self, event,
+                                                   None, event.source.parent)
+            self.lastCell = event.source.parent
+            return
+
+        # Two events are received when the caret moves
+        # to a new paragraph. The first is a focus event
+        # (in the form of object:state-changed:focused
+        # instead of focus:). The second is a caret-moved
+        # event. Just set the locusOfFocus for the first event.
+        # 
         if event.type == "object:state-changed:focused" and \
            event.source.role == rolenames.ROLE_PARAGRAPH and \
            event.source != self.currentParagraph:
             self.currentParagraph = event.source
+            orca.setLocusOfFocus(event, event.source, False)
             return
-                        
+
+        # If we get "object:state-changed:focused" events for children of 
+        # a combo-box, just set the focus to the combo box. This is needed 
+        # to help reduce the verbosity of focusing on the Calc Name combo 
+        # box (see bug #364407).
+        #
+        if event.source.parent and \
+           event.source.parent.role == rolenames.ROLE_COMBO_BOX:
+            orca.setLocusOfFocus(None, event.source.parent, False)
+            return
+
         default.Script.onStateChanged(self, event)
 
 
@@ -1321,3 +1567,166 @@ class Script(default.Script):
         string = text[max(0,startOffset):min(len(text),endOffset)]
         string = string.encode("UTF-8")
         return string
+
+    def speakCellName(self, name):
+        """Speaks the given cell name.
+
+        Arguments:
+        - name: the name of the cell
+        """
+
+        line = _("Cell ") + name
+        speech.speak(line)
+
+    def onCaretMoved(self, event):
+        """Called whenever the caret moves.
+
+        Arguments:
+        - event: the Event
+        """
+
+        # If we are FOCUSED on a paragraph inside a table cell (in Writer),
+        # then just return (modulo the special cases below). Speaking and 
+        # brailling will have been done in the onStateChanged() routine 
+        # (see bug #382415).
+        #
+        if event.source.role == rolenames.ROLE_PARAGRAPH and \
+           event.source.parent.role == rolenames.ROLE_TABLE_CELL and \
+           event.source.state.count(atspi.Accessibility.STATE_FOCUSED):
+            event_string = orca_state.lastInputEvent.event_string
+
+            # If we are moving up and down, and we are speaking-by-cell
+            # (as opposed to by-row), then speak the cell name. Otherwise
+            # just return.
+            #
+            if (event_string == "Up" or event_string == "Down"):
+                if settings.readTableCellRow == False:
+                    if event.detail1 != -1:
+                        self.speakCellName(event.source.parent.name)
+                return
+
+            # If we are moving left or right and we are in a new cell, just
+            # return.
+            #
+            if (event_string == "Left" or event_string == "Right") and \
+               self.lastCell != event.source.parent:
+                return
+
+            caretOffset = event.source.text.caretOffset
+            len = event.source.text.characterCount
+
+            # If you are in a table cell and you arrow Right, the caret 
+            # will focus at the end of the current paragraph before moving 
+            # into the next cell. To be similar to the way that caret 
+            # navigation works in other paragraphs in OOo, just return.
+            #
+            if event_string == "Right" and caretOffset == len:
+                return
+
+            # If we have moved left and the caret position is at the end of
+            # the paragraph or if we have moved right and the caret position
+            # is at the start of the text string, or the last key input was
+            # Tab or Shift-Tab, and if we are speaking-by-cell (as opposed 
+            # to by-row), then speak the cell name, otherwise just return 
+            # (see bug #382418).
+            #
+            if (event_string == "Left" and caretOffset == len) or \
+               (event_string == "Right" and caretOffset == 0) or \
+               (event_string == "Tab" or event_string == "ISO_Left_Tab"):
+                if settings.readTableCellRow == False:
+                    if event.detail1 != -1:
+                        self.speakCellName(event.source.parent.name)
+
+                # Speak a blank line, if appropriate.
+                if self.speakBlankLine(event.source):
+                    speech.speak(_("blank"), None, False)
+                return
+
+        # Speak a newline, if appropriate.
+        if self.speakNewLine(event.source):
+            speech.speak(chnames.getCharacterName("\n"), None, False)
+
+        # Speak a blank line, if appropriate.
+        if self.speakBlankLine(event.source):
+            speech.speak(_("blank"), None, False)
+
+        default.Script.onCaretMoved(self, event)
+
+
+    def speakNewLine(self, obj):
+        """Returns True if a newline should be spoken.
+           Otherwise, returns False.
+        """
+
+        # Get the the AccessibleText interrface.
+        text = obj.text
+        if not text:
+            return False
+
+        # Was a left or right-arrow key pressed?
+        if not (orca_state.lastInputEvent and \
+                orca_state.lastInputEvent.__dict__.has_key("event_string")):
+            return False
+        
+        lastKey = orca_state.lastInputEvent.event_string
+        if lastKey != "Left" and lastKey != "Right":
+            return False
+
+        # Was a control key pressed?
+        mods = orca_state.lastInputEvent.modifiers
+        isControlKey = mods & (1 << atspi.Accessibility.MODIFIER_CONTROL)
+        
+        # Get the line containing the caret
+        caretOffset = text.caretOffset
+        line = text.getTextAtOffset(caretOffset, \
+            atspi.Accessibility.TEXT_BOUNDARY_LINE_START)
+        lineStart = line[1]
+        lineEnd = line[2]
+
+        if isControlKey:  # control-right-arrow or control-left-arrow
+
+            # Get the word containing the caret.
+            word = text.getTextAtOffset(caretOffset, \
+                atspi.Accessibility.TEXT_BOUNDARY_WORD_START)
+            wordStart = word[1]
+            wordEnd = word[2]
+            
+            if lastKey == "Right":
+                if wordStart == lineStart:
+                    return True
+            else: 
+                if wordEnd == lineEnd:
+                    return True
+                
+        else:  # right arrow or left arrow
+            
+            if lastKey == "Right":
+                if caretOffset == lineStart:
+                    return True
+            else: 
+                if caretOffset == lineEnd:
+                    return True
+
+        return False
+
+
+    def speakBlankLine(self, obj):
+        """Returns True if a blank line should be spoken.
+        Otherwise, returns False.
+        """
+        
+        # Get the the AccessibleText interrface. 
+        text = obj.text
+        if not text:
+            return False
+
+        # Get the line containing the caret
+        caretOffset = text.caretOffset
+        line = text.getTextAtOffset(caretOffset, \
+            atspi.Accessibility.TEXT_BOUNDARY_LINE_START)
+
+        # If this is a blank line, announce it if the user requested
+        # that blank lines be spoken.
+        if line[1] == 0 and line[2] == 0:
+            return settings.speakBlankLines
+
