@@ -19,9 +19,9 @@
 
 """Provides a custom script for gcalctool."""
 
-__id__        = "$Id: gcalctool.py,v 1.15 2006/10/12 00:33:48 wwalker Exp $"
-__version__   = "$Revision: 1.15 $"
-__date__      = "$Date: 2006/10/12 00:33:48 $"
+__id__        = "$Id: gcalctool.py,v 1.16 2006/11/03 21:18:21 richb Exp $"
+__version__   = "$Revision: 1.16 $"
+__date__      = "$Date: 2006/11/03 21:18:21 $"
 __copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
 __license__   = "LGPL"
 
@@ -160,6 +160,7 @@ class Script(default.Script):
                                    input_event.KeyboardEvent)):
                 return
 
-            if (orca_state.lastInputEvent.event_string == "Return") \
+            if (orca_state.lastInputEvent.event_string == "space") \
+                   or (orca_state.lastInputEvent.event_string == "Return") \
                    or (orca_state.lastInputEvent.event_string == "="):
                 speech.speak(contents)
