@@ -20,12 +20,15 @@
 """Manages the default speech server for orca.  A script can use this
 as its speech server, or it can feel free to create one of its own."""
 
-__id__        = "$Id: speech.py 1312 2006-08-11 17:24:24Z wwalker $"
-__version__   = "$Revision: 1312 $"
-__date__      = "$Date: 2006-08-11 13:24:24 -0400 (Fri, 11 Aug 2006) $"
+__id__        = "$Id: speech.py,v 1.61 2006/10/13 23:39:04 wwalker Exp $"
+__version__   = "$Revision: 1.61 $"
+__date__      = "$Date: 2006/10/13 23:39:04 $"
 __copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
 __license__   = "LGPL"
 
+import logging
+log = logging.getLogger("speech")
+
 import time
 
 import debug
@@ -145,6 +148,7 @@ def speak(text, acss=None, interrupt=True):
         return
 
     debug.println(debug.LEVEL_INFO, "SPEECH OUTPUT: '" + text + "'")
+    log.info("'%s'" % text)
 
     if __speechserver:
         __speechserver.speak(text, __resolveACSS(acss), interrupt)
@@ -181,6 +185,7 @@ def speakUtterances(utterances, acss=None, interrupt=True):
     for utterance in utterances:
         debug.println(debug.LEVEL_INFO,
                       "SPEECH OUTPUT: '" + utterance + "'")
+        log.info("'%s'" % utterance)
 
     if __speechserver:
         __speechserver.speakUtterances(utterances,
