@@ -23,9 +23,9 @@ class.  For those wishing to override the speech generators, however,
 one can create a new instance and replace/extend the speech generators
 as they see fit."""
 
-__id__        = "$Id: speechgenerator.py,v 1.79 2006/10/12 00:33:47 wwalker Exp $"
-__version__   = "$Revision: 1.79 $"
-__date__      = "$Date: 2006/10/12 00:33:47 $"
+__id__        = "$Id: speechgenerator.py,v 1.88 2006/12/12 21:07:08 richb Exp $"
+__version__   = "$Revision: 1.88 $"
+__date__      = "$Date: 2006/12/12 21:07:08 $"
 __copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
 __license__   = "LGPL"
 
@@ -52,7 +52,7 @@ class SpeechGenerator:
         # script for information if we need it.
         #
         self._script = script
-        
+
         # Set up a dictionary that maps role names to functions
         # that generate speech for objects that implement that role.
         #
@@ -91,6 +91,8 @@ class SpeechGenerator:
              self._getSpeechForImage
         self.speechGenerators[rolenames.ROLE_LABEL]               = \
              self._getSpeechForLabel
+        self.speechGenerators[rolenames.ROLE_LAYERED_PANE]        = \
+             self._getSpeechForLayeredPane
         self.speechGenerators[rolenames.ROLE_LIST]                = \
              self._getSpeechForList
         self.speechGenerators[rolenames.ROLE_MENU]                = \
@@ -263,8 +265,11 @@ class SpeechGenerator:
         utterances = []
 
         if not already_focused:
-            utterances.extend(self._getSpeechForObjectLabel(obj))
-            utterances.extend(self._getSpeechForObjectName(obj))
+            label = self._getSpeechForObjectLabel(obj)
+            utterances.extend(label)
+            name = self._getSpeechForObjectName(obj)
+            if name != label:
+                utterances.extend(name)
             utterances.extend(self._getSpeechForObjectRole(obj))
 
         utterances.extend(self._getSpeechForObjectAvailability(obj))
@@ -288,9 +293,11 @@ class SpeechGenerator:
         """
 
         utterances = []
-        utterances.extend(self._getSpeechForObjectLabel(obj))
-        utterances.extend(self._getSpeechForObjectName(obj))
-        utterances.extend(self._getSpeechForObjectRole(obj))
+        label = self._getSpeechForObjectLabel(obj)
+        utterances.extend(label)
+        name = self._getSpeechForObjectName(obj)
+        if name != label:
+            utterances.extend(name)
 
         # Find all the unrelated labels in the dialog and speak them.
         #
@@ -316,9 +323,11 @@ class SpeechGenerator:
         """
 
         utterances = []
-        utterances.extend(self._getSpeechForObjectLabel(obj))
-        utterances.extend(self._getSpeechForObjectName(obj))
-        utterances.extend(self._getSpeechForObjectRole(obj))
+        label = self._getSpeechForObjectLabel(obj)
+        utterances.extend(label)
+        name = self._getSpeechForObjectName(obj)
+        if name != label:
+            utterances.extend(name)
 
         self._debugGenerator("_getSpeechForAnimation",
                              obj,
@@ -363,25 +372,23 @@ class SpeechGenerator:
 
         utterances = []
         if obj.state.count(atspi.Accessibility.STATE_CHECKED):
-            # If it's not already focused, say it's name
-            #
-            if not already_focused:
-                utterances.extend(self._getSpeechForObjectLabel(obj))
-                utterances.extend(self._getSpeechForObjectName(obj))
-                utterances.extend(self._getSpeechForObjectRole(obj))
-                utterances.append(_("checked"))
-                utterances.extend(self._getSpeechForObjectAvailability(obj))
-            else:
-                utterances.append(_("checked"))
+            checkedState = _("checked")
         else:
-            if not already_focused:
-                utterances.extend(self._getSpeechForObjectLabel(obj))
-                utterances.extend(self._getSpeechForObjectName(obj))
-                utterances.extend(self._getSpeechForObjectRole(obj))
-                utterances.append(_("not checked"))
-                utterances.extend(self._getSpeechForObjectAvailability(obj))
-            else:
-                utterances.append(_("not checked"))
+            checkedState = _("not checked")
+
+        # If it's not already focused, say it's name
+        #
+        if not already_focused:
+            label = self._getSpeechForObjectLabel(obj)
+            utterances.extend(label)
+            name = self._getSpeechForObjectName(obj)
+            if name != label:
+                utterances.extend(name)
+            utterances.extend(self._getSpeechForObjectRole(obj))
+            utterances.append(checkedState)
+            utterances.extend(self._getSpeechForObjectAvailability(obj))
+        else:
+            utterances.append(checkedState)
 
         self._debugGenerator("_getSpeechForCheckBox",
                              obj,
@@ -447,9 +454,14 @@ class SpeechGenerator:
         utterances = []
 
         if not already_focused:
-            utterances.extend(self._getSpeechForObjectLabel(obj))
+            label = self._getSpeechForObjectLabel(obj)
+            utterances.extend(label)
+        else:
+            label = None
 
-        utterances.extend(self._getSpeechForObjectName(obj))
+        name = self._getSpeechForObjectName(obj)
+        if name != label:
+            utterances.extend(name)
 
         if not already_focused:
             utterances.extend(self._getSpeechForObjectRole(obj))
@@ -603,8 +615,11 @@ class SpeechGenerator:
         # availability.]]]
         #
         utterances = []
-        utterances.extend(self._getSpeechForObjectLabel(obj))
-        utterances.extend(self._getSpeechForObjectName(obj))
+        label = self._getSpeechForObjectLabel(obj)
+        utterances.extend(label)
+        name = self._getSpeechForObjectName(obj)
+        if name != label:
+            utterances.extend(name)
 
         if obj.image:
             description = obj.image.imageDescription
@@ -659,6 +674,36 @@ class SpeechGenerator:
 
         return utterances
 
+    def _getSpeechForLayeredPane(self, obj, already_focused):
+        """Get the speech for a layered pane
+
+        Arguments:
+        - obj: the table
+        - already_focused: False if object just received focus
+
+        Returns a list of utterances to be spoken for the object.
+        """
+
+        utterances = self._getDefaultSpeech(obj, already_focused)
+
+        self._debugGenerator("_getSpeechForLayeredPane",
+                             obj,
+                             already_focused,
+                             utterances)
+
+        # If this has no children, then let the user know.
+        #
+        hasItems = False
+        for i in range(0, obj.childCount):
+            child = obj.child(i)
+            if child.state.count(atspi.Accessibility.STATE_SHOWING):
+                hasItems = True
+                break
+        if not hasItems:
+            utterances.append(_("0 items"))
+
+        return utterances
+
     def _getSpeechForList(self, obj, already_focused):
         """Get the speech for a list.
 
@@ -767,9 +812,10 @@ class SpeechGenerator:
         if len(utterances) == 0:
             if obj.name and (len(obj.name)):
                 utterances.append(obj.name)
-        utterances.extend(self._getSpeechForObjectRole(obj))
+        if obj.role != rolenames.ROLE_PARAGRAPH:
+            utterances.extend(self._getSpeechForObjectRole(obj))
 
-        [text, startOffset, endOffset] = util.getTextLineAtCaret(obj)
+        [text, caretOffset, startOffset] = util.getTextLineAtCaret(obj)
         utterances.append(text)
 
         self._debugGenerator("_getSpeechForText",
@@ -858,8 +904,11 @@ class SpeechGenerator:
         utterances = []
 
         if not already_focused:
-            utterances.extend(self._getSpeechForObjectLabel(obj))
-            utterances.extend(self._getSpeechForObjectName(obj))
+            label = self._getSpeechForObjectLabel(obj)
+            utterances.extend(label)
+            name = self._getSpeechForObjectName(obj)
+            if name != label:
+                utterances.extend(name)
             utterances.extend(self._getSpeechForObjectRole(obj))
 
         utterances.append(percentage)
@@ -903,29 +952,22 @@ class SpeechGenerator:
 
         utterances = []
         if obj.state.count(atspi.Accessibility.STATE_CHECKED):
-            # If it's not already focused, say it's name
-            #
-            if not already_focused:
-                # The label is handled as a context in default.py
-                #
-                #utterances.extend(self._getSpeechForObjectLabel(obj))
-                utterances.extend(self._getSpeechForObjectName(obj))
-                utterances.extend(self._getSpeechForObjectRole(obj))
-                utterances.append(_("selected"))
-                utterances.extend(self._getSpeechForObjectAvailability(obj))
-            else:
-                utterances.append(_("selected"))
+            selectionState = _("selected")
         else:
-            if not already_focused:
-                # The label is handled as a context in default.py
-                #
-                #utterances.extend(self._getSpeechForObjectLabel(obj))
-                utterances.extend(self._getSpeechForObjectName(obj))
-                utterances.extend(self._getSpeechForObjectRole(obj))
-                utterances.append(_("not selected"))
-                utterances.extend(self._getSpeechForObjectAvailability(obj))
-            else:
-                utterances.append(_("not selected"))
+            selectionState = _("not selected")
+
+        # If it's not already focused, say it's name
+        #
+        if not already_focused:
+            # The label is handled as a context in default.py
+            #
+            #utterances.extend(self._getSpeechForObjectLabel(obj))
+            utterances.extend(self._getSpeechForObjectName(obj))
+            utterances.append(selectionState)
+            utterances.extend(self._getSpeechForObjectRole(obj))
+            utterances.extend(self._getSpeechForObjectAvailability(obj))
+        else:
+            utterances.append(selectionState)
 
         self._debugGenerator("_getSpeechForRadioButton",
                              obj,
@@ -1024,6 +1066,9 @@ class SpeechGenerator:
         try:
             minimumIncrement = value.minimumIncrement
         except:
+            pass
+
+        if minimumIncrement == 0.0:
             minimumIncrement = (value.maximumValue - value.minimumValue) \
                                / 100.0
 
@@ -1132,6 +1177,17 @@ class SpeechGenerator:
                              already_focused,
                              utterances)
 
+        # If this is a table with no children, then let the user know.
+        #
+        hasItems = False
+        for i in range(0, obj.childCount):
+            child = obj.child(i)
+            if child.state.count(atspi.Accessibility.STATE_SHOWING):
+                hasItems = True
+                break
+        if not hasItems:
+            utterances.append(_("0 items"))
+
         return utterances
 
     def _getSpeechForTableCell(self, obj, already_focused):
@@ -1176,6 +1232,12 @@ class SpeechGenerator:
             else:
                 utterances.append(_("collapsed"))
 
+            # If this is an expandable table cell with no children, then
+            # let the user know.
+            #
+            if obj.childCount == 0:
+                utterances.append(_("0 items"))
+
         self._debugGenerator("_getSpeechForTableCell",
                              obj,
                              already_focused,
@@ -1221,13 +1283,15 @@ class SpeechGenerator:
                         showing = cell.state.count( \
                                           atspi.Accessibility.STATE_SHOWING)
                         if showing:
-                            utterances.extend(self._getSpeechForTableCell(cell, 
+                            utterances.extend(self._getSpeechForTableCell(cell,
                                                            already_focused))
                 else:
                     utterances.extend(self._getSpeechForTableCell(obj,
                                                            already_focused))
             else:
                 utterances = self._getSpeechForTableCell(obj, already_focused)
+        else:
+            utterances = self._getSpeechForTableCell(obj, already_focused)
 
         return utterances
 
@@ -1330,25 +1394,23 @@ class SpeechGenerator:
 
         utterances = []
         if obj.state.count(atspi.Accessibility.STATE_CHECKED):
-            # If it's not already focused, say it's name
-            #
-            if not already_focused:
-                utterances.extend(self._getSpeechForObjectLabel(obj))
-                utterances.extend(self._getSpeechForObjectName(obj))
-                utterances.extend(self._getSpeechForObjectRole(obj))
-                utterances.append(_("pressed"))
-                utterances.extend(self._getSpeechForObjectAvailability(obj))
-            else:
-                utterances.append(_("pressed"))
+            checkedState = _("pressed")
         else:
-            if not already_focused:
-                utterances.extend(self._getSpeechForObjectLabel(obj))
-                utterances.extend(self._getSpeechForObjectName(obj))
-                utterances.extend(self._getSpeechForObjectRole(obj))
-                utterances.append(_("not pressed"))
-                utterances.extend(self._getSpeechForObjectAvailability(obj))
-            else:
-                utterances.append(_("not pressed"))
+            checkedState = _("not pressed")
+
+        # If it's not already focused, say it's name
+        #
+        if not already_focused:
+            label = self._getSpeechForObjectLabel(obj)
+            utterances.extend(label)
+            name = self._getSpeechForObjectName(obj)
+            if name != label:
+                utterances.extend(name)
+            utterances.extend(self._getSpeechForObjectRole(obj))
+            utterances.append(checkedState)
+            utterances.extend(self._getSpeechForObjectAvailability(obj))
+        else:
+            utterances.append(checkedState)
 
         self._debugGenerator("_getSpeechForToggleButton",
                              obj,
@@ -1495,6 +1557,7 @@ class SpeechGenerator:
             if parent == stopAncestor:
                 break
             if (parent.role != rolenames.ROLE_FILLER) \
+                and (parent.role != rolenames.ROLE_SECTION) \
                 and (parent.role != rolenames.ROLE_LAYERED_PANE) \
                 and (parent.role != rolenames.ROLE_SPLIT_PANE) \
                 and (parent.role != rolenames.ROLE_SCROLL_PANE) \
@@ -1539,4 +1602,4 @@ class SpeechGenerator:
 
         utterances.reverse()
 
-        return [" ".join(utterances)]
+        return utterances
